"use strict";
/**
 * @license
 * Copyright (c) 2019 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt The complete set of authors may be found
 * at http://polymer.github.io/AUTHORS.txt The complete set of contributors may
 * be found at http://polymer.github.io/CONTRIBUTORS.txt Code distributed by
 * Google as part of the polymer project is also subject to an additional IP
 * rights grant found at http://polymer.github.io/PATENTS.txt
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeBackSchemaIfNeeded = exports.parseConfigFile = void 0;
const fsExtra = __importStar(require("fs-extra"));
const jsonschema = __importStar(require("jsonschema"));
const browser_1 = require("./browser");
const config_1 = require("./config");
const defaults = __importStar(require("./defaults"));
const types_1 = require("./types");
const util_1 = require("./util");
/**
 * Validate the given JSON object parsed from a config file, and expand it into
 * a fully specified configuration.
 */
async function parseConfigFile(parsedJson) {
    const schema = require('../config.schema.json');
    const result = jsonschema.validate(parsedJson, schema, { propertyName: 'config' });
    if (result.errors.length > 0) {
        throw new Error([...new Set(result.errors.map(customizeJsonSchemaError))].join('\n'));
    }
    const validated = parsedJson;
    const root = validated.root || '.';
    const benchmarks = [];
    for (const benchmark of validated.benchmarks) {
        for (const expanded of applyExpansions(benchmark)) {
            benchmarks.push(applyDefaults(await parseBenchmark(expanded, root)));
        }
    }
    return {
        root,
        sampleSize: validated.sampleSize,
        timeout: validated.timeout,
        horizons: validated.horizons !== undefined ?
            config_1.parseHorizons(validated.horizons) :
            undefined,
        benchmarks,
        resolveBareModules: validated.resolveBareModules,
    };
}
exports.parseConfigFile = parseConfigFile;
/**
 * Some of the automatically generated jsonschema errors are unclear, e.g. when
 * there is a union of complex types they are reported as "[schema1],
 * [schema2]" etc.
 */
function customizeJsonSchemaError(error) {
    if (error.property.match(/^config\.benchmarks\[\d+\]\.measurement$/)) {
        return `${error.property} is not one of: ${[...types_1.measurements].join(', ')}` +
            ' or an object like `performanceEntry: string`';
    }
    return error.toString();
}
async function parseBenchmark(benchmark, root) {
    const spec = {};
    if (benchmark.name !== undefined) {
        spec.name = benchmark.name;
    }
    if (benchmark.browser !== undefined) {
        let browser;
        if (typeof benchmark.browser === 'string') {
            browser = Object.assign(Object.assign({}, browser_1.parseBrowserConfigString(benchmark.browser)), { windowSize: {
                    width: defaults.windowWidth,
                    height: defaults.windowHeight,
                } });
        }
        else {
            browser = parseBrowserObject(benchmark.browser);
        }
        browser_1.validateBrowserConfig(browser);
        spec.browser = browser;
    }
    if (benchmark.measurement === 'callback') {
        spec.measurement = [{
                mode: 'callback',
            }];
    }
    else if (benchmark.measurement === 'fcp') {
        spec.measurement = [{
                mode: 'performance',
                entryName: 'first-contentful-paint',
            }];
    }
    else if (benchmark.measurement === 'global') {
        spec.measurement = [{
                mode: 'expression',
                expression: benchmark.measurementExpression || defaults.measurementExpression,
            }];
    }
    else if (Array.isArray(benchmark.measurement)) {
        spec.measurement = benchmark.measurement;
    }
    else if (benchmark.measurement !== undefined) {
        spec.measurement = [benchmark.measurement];
    }
    const url = benchmark.url;
    if (url !== undefined) {
        if (util_1.isHttpUrl(url)) {
            spec.url = {
                kind: 'remote',
                url,
            };
        }
        else {
            let urlPath, queryString;
            const q = url.indexOf('?');
            if (q !== -1) {
                urlPath = url.substring(0, q);
                queryString = url.substring(q);
            }
            else {
                urlPath = url;
                queryString = '';
            }
            spec.url = {
                kind: 'local',
                urlPath: await config_1.urlFromLocalPath(root, urlPath),
                queryString,
            };
            if (benchmark.packageVersions !== undefined) {
                spec.url.version = {
                    label: benchmark.packageVersions.label,
                    dependencyOverrides: benchmark.packageVersions.dependencies,
                };
            }
        }
    }
    return spec;
}
function parseBrowserObject(config) {
    const parsed = {
        name: config.name,
        headless: ('headless' in config && config.headless) || defaults.headless,
        windowSize: ('windowSize' in config && config.windowSize) || {
            width: defaults.windowWidth,
            height: defaults.windowHeight,
        },
    };
    if ('cpuThrottlingRate' in config) {
        parsed.cpuThrottlingRate = config.cpuThrottlingRate;
    }
    if (config.remoteUrl) {
        parsed.remoteUrl = config.remoteUrl;
    }
    if ('binary' in config && config.binary) {
        parsed.binary = config.binary;
    }
    if ('addArguments' in config && config.addArguments) {
        parsed.addArguments = config.addArguments;
    }
    if ('removeArguments' in config && config.removeArguments) {
        parsed.removeArguments = config.removeArguments;
    }
    if ('preferences' in config && config.preferences) {
        parsed.preferences = config.preferences;
    }
    return parsed;
}
function applyExpansions(bench) {
    if (bench.expand === undefined || bench.expand.length === 0) {
        return [bench];
    }
    const expanded = [];
    for (const expansion of bench.expand) {
        for (const expandedBench of applyExpansions(expansion)) {
            expanded.push(Object.assign(Object.assign({}, bench), expandedBench));
        }
    }
    return expanded;
}
function applyDefaults(partialSpec) {
    const url = partialSpec.url;
    let { name, measurement, browser } = partialSpec;
    if (url === undefined) {
        // Note we can't validate this with jsonschema, because we only need to
        // ensure we have a URL after recursive expansion; so at any given level
        // the URL could be optional.
        throw new Error('No URL specified');
    }
    if (url.kind === 'remote') {
        if (name === undefined) {
            name = url.url;
        }
    }
    else {
        if (name === undefined) {
            name = url.urlPath + url.queryString;
        }
    }
    if (browser === undefined) {
        browser = {
            name: defaults.browserName,
            headless: defaults.headless,
            windowSize: {
                width: defaults.windowWidth,
                height: defaults.windowHeight,
            },
        };
    }
    if (measurement === undefined) {
        measurement = [defaults.measurement(url)];
    }
    return { name, url, browser, measurement };
}
async function writeBackSchemaIfNeeded(rawConfigObj, configFile) {
    // Add the $schema field to the original config file if it's absent.
    // We only want to do this if the file validated though, so we don't mutate
    // a file that's not actually a tachometer config file.
    if (!('$schema' in rawConfigObj)) {
        const $schema = 'https://raw.githubusercontent.com/Polymer/tachometer/master/config.schema.json';
        // Extra IDE features can be activated if the config file has a schema.
        const withSchema = Object.assign({ $schema }, rawConfigObj);
        const contents = JSON.stringify(withSchema, null, 2);
        await fsExtra.writeFile(configFile, contents, { encoding: 'utf-8' });
    }
}
exports.writeBackSchemaIfNeeded = writeBackSchemaIfNeeded;
//# sourceMappingURL=configfile.js.map